self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "ce435d59eeb0ea7a3a045313d8f0e1ef",
    "url": "./index.html"
  },
  {
    "revision": "25688505d534eab23599",
    "url": "./static/js/2.23f7626a.chunk.js"
  },
  {
    "revision": "05316acded932624393d390915caa64d",
    "url": "./static/js/2.23f7626a.chunk.js.LICENSE.txt"
  },
  {
    "revision": "bdb2c39fcf7746634384",
    "url": "./static/js/main.3cc30e2c.chunk.js"
  },
  {
    "revision": "c6c613ae59ef7502cd97",
    "url": "./static/js/runtime-main.3ca778a4.js"
  },
  {
    "revision": "9cfe50b68736f61d061af526da6aaf14",
    "url": "./static/media/bezier.9cfe50b6.cjs"
  }
]);